from ..core.runnable import Retriever, MarkMeta
from ..community.faiss import FaissDB
from ..community.dashscope.reranker import DashScopeReranker

__all__ = [
    "FaissDB",
    "Retriever",
    "MarkMeta",
    "DashScopeReranker",
]



from typing import Union, List, Callable

from ......io import EventBlock
from ....prompt_template import PromptTemplate
from ...base import BaseAgent
from ..base import BaseTeam

class ReAct(BaseTeam):
    """
    ReAct 主要通过 CoT 思想来控制提示语生成，在循环过程中构建「思考-行动-观察」的循环，直至结束。
    """
    def __init__(self, leader: BaseAgent, leader_prompt_template: str="", condition: Callable=None, **kwargs):
        self.leader = leader()
        self.prompt_template = PromptTemplate(leader_prompt_template or "TEAM/ReAct")
        self.condition = condition or (lambda x: True)

        super().__init__(leader=leader, **kwargs)
    
    def call(self, prompt: Union[str, List[dict]], **kwargs):
        """
        调用团队协作完成任务。
        """ 
        self._completed_teamwork = []
        self._last_output = None
        max_rounds = 10
        task_index = 0
        for round in range(max_rounds):
            to_continue = False
            yield EventBlock("agent", f"ROUND {round + 1} @{self.leader.name}")
            yield from self.leader.call(prompt, new_chat=True, **kwargs)
            self._last_output = self.leader.last_output
            if not self._completed_teamwork:
                self._completed_teamwork.append(f'[总任务目标] {self.leader.provider_dict.get("task")}')

            self_solve = self.extract_self_solve(self.leader.last_output)
            if self_solve:
                self._completed_teamwork.append(f'[行动] 该子任务自己解决，结果为 {self_solve}')

            tasks = self.extract_task_dispatch(self.leader.last_output)
            for task in tasks:
                task_detail = task["task_detail"]
                member = self.find_member(task["member_name"])
                if member and task_detail:
                    self._completed_teamwork.append(f'[行动] #E{task_index + 1} 子任务交给 @{member.name}, 要求为 {task_detail}')
                    yield EventBlock("agent", f"ROUND {round + 1} @{member.name}")
                    yield from member.call(task_detail, new_chat=True, **kwargs)

                    self._completed_teamwork.append(f'[观察] @{member.name} 已完成 #E{task_index + 1} 子任务，回复内容为 {member.last_output}')
                    to_continue = True

            if not to_continue:
                break

        final_answers = self.extract_answers(self.leader.last_output)
        if final_answers:
            self._completed_teamwork.append(f'[最终答案] \n{final_answers[-1]}')
            self._last_output = final_answers[-1]
            yield EventBlock("final_answer", self._last_output)
        else:
            yield EventBlock("final_answer", "经过多次尝试，没有找到最终答案")


import pandas as pd
from typing import Union

class Dataset:
    def __init__(self, df: Union[pd.DataFrame] = None, desc: str = None):
        self.df = df
        self.desc = desc
    
    def __str__(self):
        return self.desc

    def __repr__(self):
        return f"Dataset(desc={self.desc})"



from typing import List, Union

from .....io import EventBlock
from .....utils import compress_text
from ..chat import ChatAgent

class Discuss():
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def run(self, outline: str, *args, **kwargs):
        pass


from ..community.zhipu import ZhipuEmbeddings
from ..community.dashscope import TextEmbeddings
from ..community.openai import OpenAIEmbeddings
from ..community.hugging_face import HuggingFaceEmbeddings

__all__ = [
    "ZhipuEmbeddings",
    "TextEmbeddings",
    "OpenAIEmbeddings",
    "HuggingFaceEmbeddings",
]



import re
from typing import Any, Set, Union, List, Dict
from ...document import Document
from ..vectordb import VectorDB
from ..message import Message

class ResourceManager:
    @classmethod
    def available_init_params(cls):
        """
        返回当前可用的参数列表。
        """
        return {
            "resources": "资源列表",
        }

    def __init__(self, resources: List[Any] = None):
        """
        管理智能体使用到的图片、视频、音频、数据、文档等资源。
        """
        self.resources = []
        for r in (resources or []):
            self.add_resource(r)

    def add_resource(self, *content: Union[str, List[Any], Dict[str, str]]):
        for c in content:
            if isinstance(c, str): 
                self.resources.append(Message(role="resource", content=[{"text": c}]))
            elif isinstance(c, dict):
                self.resources.append(Message(role="resource", content=[{k:v} for k, v in c.items()]))
            elif isinstance(c, list):
                self.resources.append(Message(role="resource", content=c))
            else:
                raise ValueError("Resource must be a string or a dict")

    def add_images(self, description: str, images: List[str]):
        if not isinstance(images, list):
            images = [images]
        new_rc = Message(role="resource", content=[{"image": image} for image in images] + [{"text": description}])
        if self.get_signature(new_rc) in [self.get_signature(r) for r in self.resources]:
            return None
        else:
            self.resources.append(new_rc)
            return new_rc

    def get_signature(self, message: Message):
        texts = []
        for x in message.content:
            for v in x.values():
                texts.append(v)
        return "__".join(texts)

    def to_messages(self, style: str="qwen_vl"):
        return [r.to_dict(style=style) for r in self.resources]

    def clear_resources(self):
        self.resources.clear()

    def get_text(self):
        return [r.to_dict(style="text")['content'] for r in self.resources]
    
    def get_resources(self, query: str=None):
        """
        获取资源描述。
        """
        resources = []
        for r in self.to_messages():
            if r['role'] == "resource":
                text_descriptions = []
                images = []
                audios = []
                videos = []
                for items in r['content']:
                    for k, v in items.items():
                        if k == 'text':
                            text_descriptions.append(v)
                        elif k == 'image':
                            images.append(v)
                        elif k == 'audio':
                            audios.append(v)
                        elif k == 'video':
                            videos.append(v)
                desc = "resource: "
                if len(images) > 0:
                    desc += "[图片] " + " ".join(images) + ", "
                if len(audios) > 0:
                    desc += "[音频] " + " ".join(audios) + ", "
                if len(videos) > 0:
                    desc += "[视频] " + " ".join(videos) + ", "
                if len(text_descriptions) > 0:
                    desc += ", ".join(text_descriptions)
                resources.append(desc)
        return resources



import json
import copy
import asyncio

from abc import abstractmethod
from typing import Union, List, Dict, Any, Set, Callable

from .....utils import merge_tool_calls, extract_text, raise_invalid_params, filter_kwargs
from .....io import EventBlock, EndBlock, NewLineBlock
from ...base import Runnable
from ...message import Messages
from ..base import BaseAgent
from ..knowledge_manager import KnowledgeManager
from .tools_calling import BaseToolCalling, OpenAIToolsCalling
from .tools_manager import ToolsManager
from .memory_manager import MemoryManager

class ChatAgent(BaseAgent, KnowledgeManager, MemoryManager, ToolsManager):
    """
    对话智能体是基于大模型实现的智能体，可以用于对话生成、对话理解等场景。
    """
    @classmethod
    def available_init_params(cls):
        """
        返回当前可用的参数列表。
        """
        return {
            "end_chk": "是否在最后输出一个 EndBlock",
            "start_marker": "开始标记，默认为 ```",
            "end_marker": "结束标记，默认为 ```",
            **BaseAgent.available_init_params(),
            **KnowledgeManager.available_init_params(),
            **ToolsManager.available_init_params(),
            **MemoryManager.available_init_params(),
        }

    def __init__(
        self,
        end_chk: bool = False,
        start_marker: str=None,
        end_marker: str=None,
        **kwargs
    ):
        """
        对话智能体支持的核心能力包括：
        - tools：工具调用
        - memory：记忆管理
        - knowledge：知识管理
        """
        raise_invalid_params(kwargs, self.__class__.available_init_params())

        kwargs["tool_params"] = kwargs.get("tool_params", {"prompt": "详细描述用户问题"})
        BaseAgent.__init__(self, **filter_kwargs(kwargs, BaseAgent.available_init_params()))
        KnowledgeManager.__init__(self, **filter_kwargs(kwargs, KnowledgeManager.available_init_params()))
        ToolsManager.__init__(self, **filter_kwargs(kwargs, ToolsManager.available_init_params()))

        self.end_chk = end_chk

        self.start_marker = start_marker or "```"
        self.end_marker = end_marker or "```"

        # 在子类中应当将模型参数保存到这个属性中，以便持久化管理
        self.model_args = {"base_url": None, "api_key": None}
        self.default_call_args = {"model": None}

        # 增加的可绑定变量
        self._task = ""
        self._tools_to_exec = self.get_tools()
        self._resources = ""

        MemoryManager.__init__(self, **filter_kwargs(kwargs, MemoryManager.available_init_params()))
    
    @property
    def runnable_info(self):
        info = super().runnable_info
        info.update({
            "model_name": self.default_call_args.get("model"),
            **self.model_args
        })
        return info

    @property
    def tools_calling_steps(self):
        steps = []
        for h in self.tools_handlers:
            steps.extend(h.steps)
        return steps

    @property
    def provider_dict(self):
        local_dict = {
            "task": self._task,
            "tools_name": ",".join([a.name for a in self._tools_to_exec]),
            "tools_desc": "\n".join(json.dumps(t.tool_desc, ensure_ascii=False) for t in self._tools_to_exec),
            "knowledge": self.get_knowledge(self._task),
            "resources": self.get_resources(self._task),
        }
        return {
            **super().provider_dict,
            **{k:v for k,v in local_dict.items() if v is not None},
        }

    @abstractmethod
    def generate(self, prompt: Union[str, List[dict]], *args, **kwargs):
        raise NotImplementedError("子类必须实现 generate 方法")
    
    async def async_generate(self, prompt: Union[str, List[dict]], *args, **kwargs):
        loop = asyncio.get_running_loop()
        for block in await self.run_in_executor(self.generate, prompt, *args, **kwargs):
            yield block

    def call(self, prompt: Any, *args, **kwargs):
        # 兼容 Runnable 类型，将其上一次的输出作为 prompt 输入
        if isinstance(prompt, Runnable):
            prompt = prompt.last_output

        new_chat = kwargs.pop("new_chat", False)

        messages_std = Messages(prompt, style="text")
        self._task = messages_std[-1].content

        # 根据模板中是否直接使用 tools_desc 来替换 tools 参数
        self._tools_to_exec = self.get_tools(kwargs.get("tools", []))
        for tool in self._tools_to_exec:
            self.bind_consumer(tool, dynamic=True)
        if "tools_desc" in self.get_bound_vars(Messages(prompt), new_chat=new_chat):
            kwargs["tools"] = None
        else:
            kwargs["tools"] = self.get_tools_desc(kwargs.get("tools", []))

        # 重新绑定工具处理的 handlers
        for h in self.tools_handlers:
            h.reset(self._tools_to_exec)

        remember_rounds = kwargs.pop("remember_rounds", self.remember_rounds)
        yield EventBlock("info", f'记住 {remember_rounds} 轮对话')

        chat_memory = self.build_chat_memory(
            prompt=prompt, # 这里依然使用 prompt，而不是 messages_std，因为需要判断到底角色是 system 还是 user
            new_chat=new_chat,
            remember_rounds=remember_rounds
        )

        to_continue_call_llm = True
        while to_continue_call_llm:
            to_continue_call_llm = False

            # 执行模型生成任务
            output_text, tools_call = "", []
            for block in self.generate(chat_memory, *args, **kwargs):
                yield block
                if block.block_type == "chunk":
                    output_text += block.text
                elif block.block_type == "text_final":
                    output_text = block.text
                elif block.block_type == "tools_call_chunk":
                    tools_call.append(json.loads(block.text))

            openai_tools_calling_steps = merge_tool_calls(tools_call)
            if openai_tools_calling_steps:
                # 从返回参数中解析工具
                handler_openai = OpenAIToolsCalling(tools_to_exec=self._tools_to_exec)
                # 处理在返回结构中包含的 openai 风格的 tools-calling 工具调用，包括将结果追加到记忆中
                for block in handler_openai.handle(openai_tools_calling_steps, chat_memory, self.memory):
                    if isinstance(block, EventBlock) and block.block_type == "tool_resp_final":
                        to_continue_call_llm = True
                    yield block
            else:
                final_output_text = extract_text(output_text, self.start_marker, self.end_marker)
                chat_memory.append({
                    "role": "assistant",
                    "content": final_output_text
                })
                self.remember_response(final_output_text)
                yield EventBlock("text_final", final_output_text)

                # 将最终的输出结果保存到 last_output 属性中
                self._last_output = final_output_text

                # 从文本中解析工具
                if "parse" in self.tools_behavior:
                    for handler in self.tools_handlers:
                        steps = handler.extract_tools_call(final_output_text)
                        if steps and "execute" in self.tools_behavior:
                            for block in handler.handle(
                                steps,
                                short_term_memory=chat_memory,
                                long_term_memory=self.memory
                            ):
                                is_final_block = isinstance(block, EventBlock) and block.block_type == "tool_resp_final"
                                if "continue" in self.tools_behavior and is_final_block:
                                    to_continue_call_llm = True
                                yield block

        if self.end_chk:
            yield EndBlock(self.last_output)

    async def async_call(self, prompt: Any, *args, **kwargs):
        # 兼容 Runnable 类型，将其上一次的输出作为 prompt 输入
        if isinstance(prompt, Runnable):
            prompt = prompt.last_output

        new_chat = kwargs.pop("new_chat", False)

        messages_std = Messages(prompt, style="text")
        self._task = messages_std.messages[-1].content

        for tool in self._tools_to_exec:
            self.bind_consumer(tool, dynamic=True)

        # 根据模板中是否直接使用 tools_desc 来替换 tools 参数
        self._tools_to_exec = self.get_tools(kwargs.get("tools", []))
        if "tools_desc" in self.get_bound_vars(Messages(prompt), new_chat=new_chat):
            kwargs["tools"] = None
        else:
            kwargs["tools"] = self.get_tools_desc(kwargs.get("tools", []))

        # 重新绑定工具处理的 handlers
        for h in self.tools_handlers:
            h.reset(self._tools_to_exec)

        remember_rounds = kwargs.pop("remember_rounds", self.remember_rounds)
        yield EventBlock("info", f'记住 {remember_rounds} 轮对话')

        chat_memory = self.build_chat_memory(
            prompt=prompt,
            new_chat=new_chat,
            remember_rounds=remember_rounds
        )

        to_continue_call_llm = True
        while to_continue_call_llm:
            to_continue_call_llm = False
            output_text, tools_call = "", []

            async for block in self.async_generate(chat_memory, *args, **kwargs):
                yield block
                if block.block_type == "chunk":
                    output_text += block.text
                elif block.block_type == "text_final":
                    output_text = block.text
                elif block.block_type == "tools_call_chunk":
                    tools_call.append(json.loads(block.text))

            final_tools_call = merge_tool_calls(tools_call)
            if final_tools_call:
                handler_openai = OpenAIToolsCalling(
                    short_term_memory=chat_memory,
                    long_term_memory=self.memory,
                    tools_to_exec=self._tools_to_exec
                )
                async for block in handler_openai.async_handle(final_tools_call):
                    if isinstance(block, EventBlock) and block.block_type == "tool_resp_final":
                        to_continue_call_llm = True
                    yield block
            else:
                final_output_text = extract_text(output_text, self.start_marker, self.end_marker)
                chat_memory.append({
                    "role": "assistant",
                    "content": final_output_text
                })
                self.remember_response(final_output_text)
                yield EventBlock("text_final", final_output_text)

                if "parse" in self.tools_behavior:
                    for handler in self.tools_handlers:
                        steps = handler.extract_tools_call(final_output_text)
                        if steps and "execute" in self.tools_behavior:
                            for block in handler.handle(
                                steps,
                                short_term_memory=chat_memory,
                                long_term_memory=self.memory
                            ):
                                is_final_block = isinstance(block, EventBlock) and block.block_type == "tool_resp_final"
                                if "continue" in self.tools_behavior and is_final_block:
                                    to_continue_call_llm = True
                                yield block

        if self.end_chk:
            yield EndBlock(self.last_output)



import json
from typing import AsyncIterable, Union
import logging
import asyncio

from .block import EventBlock

async def event_stream(runnable: "Runnable", *args, **kwargs):
    """
    生成适合于 Web 处理的 SSE 事件流格式的数据。
    """

    resp = runnable.async_call(*args, **kwargs)
    async for block in resp:
        if isinstance(block, EventBlock):
            yield f"event: {block.block_type}\ndata: {block.text}\n\n"
        elif isinstance(block, str):
            yield f"data: {block}\n\n"
        else:
            # logging.error(f"Unknown block type: {block}")
            # 在生产环境中，可能不想中断整个流，而是跳过无效的块
            continue
        await asyncio.sleep(0)



import os

def get_folder_root():
    return get_env("ILLUFLY_ROOT")

def get_env(key: str=None):
    """
    环境变量的默认值。
    """
    default_values = {
        # 文件夹配置
        "ILLUFLY_ROOT": "./",
        "ILLUFLY_PROMPT_TEMPLATE_LOCAL_FOLDER": "__PROMPTS__",
        "ILLUFLY_DOCS": "__DOCS__",
        "ILLUFLY_MEDIA": "__MEDIA__",
        "ILLUFLY_CACHE_EMBEDDINGS": "__CACHE_EMBEDDINGS__",
        "ILLUFLY_UPLOAD_CACHE_DIR": "__UPLOAD_CACHE__",

        # 结束语
        "ILLUFLY_AIGC_INFO_DECLARE": "内容由AI生成，其观点仅代表创作者个人立场",
        "ILLUFLY_AIGC_INFO_CHK": "可联系服务商查验校验码",

        # 扩写标签
        "ILLUFLY_OUTLINE_START": "<OUTLINE>",
        "ILLUFLY_OUTLINE_END": "</OUTLINE>",

        # 颜色
        "ILLUFLY_COLOR_DEFAULT": "黑色",
        "ILLUFLY_COLOR_INFO": "蓝色",
        "ILLUFLY_COLOR_TEXT": "黄色",
        "ILLUFLY_COLOR_WARN": "红色",
        "ILLUFLY_COLOR_CHUNK": "绿色",
        "ILLUFLY_COLOR_FINAL": "青色",
        "ILLUFLY_COLOR_FRONT_MATTER": "品红",

        # HTTP
        "HTTP_CHECK_RESULT_SECONDS": 2,

        # DashScope
        "DASHSCOPE_BASE_URL": "https://dashscope.aliyuncs.com/api/v1"
    }
    if key:
        if key not in default_values:
            raise ValueError(f"Environ Value [{key}] Not Exist !!!")
        else:
            return os.getenv(key) or default_values[key]
    else:
        return default_values

def color_code(color_name: str):
    colors = {
        "黑色": "\033[30m",
        "红色": "\033[31m",
        "绿色": "\033[32m",
        "黄色": "\033[33m",
        "蓝色": "\033[34m",
        "品红": "\033[35m",
        "青色": "\033[36m",
        "白色": "\033[37m",
        "重置": "\033[0m" 
    }
    return colors.get(color_name, '黑色')




import json
from typing import List, Dict, Any

def merge_json_blocks(blocks: List[Dict[str, Any]]) -> Dict[str, Any]:
    merged_result = {}

    for block in blocks:
        for key, value in block.items():
            if key not in merged_result:
                merged_result[key] = value
            else:
                if isinstance(value, dict) and isinstance(merged_result[key], dict):
                    merged_result[key] = merge_json_blocks([merged_result[key], value])
                elif key not in ['type'] and isinstance(value, str) and isinstance(merged_result[key], str):
                    merged_result[key] += value
                elif value != merged_result[key]:
                    merged_result[key] = value

    return merged_result

def merge_blocks_by_index(blocks: List[Dict[str, Any]]) -> Dict[int, Dict[str, Any]]:
    index_groups = {}
    for block in blocks:
        index = block.get('index')
        if index is not None:
            if index not in index_groups:
                index_groups[index] = []
            index_groups[index].append(block)
    
    merged_results = {index: merge_json_blocks(group) for index, group in index_groups.items()}
    return merged_results



from .base import Runnable
from .selector import Selector
from .agent import BaseAgent, ChatAgent, ChatPool, Retriever
from .agent.flow import  ReAct, ReWOO
from .agent.chat.tools_calling import BaseToolCalling, ToolCall, SubTask, Plans
from .prompt_template import PromptTemplate
from .importer import Importer
from .embeddings import BaseEmbeddings
from .vectordb import VectorDB
from .markmeta import MarkMeta
from .reranker import BaseReranker



from ..community.openai import ChatOpenAI
from ..community.zhipu import ChatZhipu
from ..community.dashscope import ChatQwen, ChatQwenVL

from ..core.runnable.agent import ChatPool, FlowAgent, ReAct, ReWOO
from ..core.runnable import Selector

from .fake import FakeLLM

__all__ = [
    "Selector",
    "FlowAgent",
    "ReAct",
    "ReWOO",
    "ChatPool",
    "ChatOpenAI",
    "ChatZhipu",
    "ChatQwen",
    "FakeLLM",
]



import asyncio
import inspect

from typing import Union, List, Dict, Any, Callable, Generator, AsyncGenerator
from abc import ABC, abstractmethod
from functools import partial

from .executor_manager import ExecutorManager
from .binding_manager import BindingManager
from ...io import log, EventBlock
from ...utils import filter_kwargs, raise_invalid_params

class Runnable(ABC, ExecutorManager, BindingManager):
    """
    实现基本可运行类，定义了可运行的基本接口。
    只要继承该类，就可以作为智能体的工具使用。

    实现机制：
    - 支持 EventBlock 流式输出句法
    - 实现 __call__ 方法，来简化流输出调用
    - 支持 call 同步方法调用
    - 支持 async_call 方法调用，并在 Runnable 中已实现默认版本
    - 支持 绑定机制

    什么时候直接从 Runnable 继承？
    - 如果需要支持流式输出
    - 仅仅做数据处理，而没有过多服务端调度逻辑

    什么时候转而使用BaseAgent？
    - 如果需要在多智能体中协同处理
    - 如果包含服务端调度逻辑

    """
    @classmethod
    def available_init_params(cls):
        """
        返回当前可用的参数列表。
        """
        return {
            "name": "Runnable 名称，默认为 {类名}.{id}",
            "handlers": "EventBlock 迭代器处理函数列表，默认为 [log]，当调用 call 方法时，会使用该列表中的函数逐个处理 EventBlock",
            **ExecutorManager.available_init_params(),
            **BindingManager.available_init_params(),
        }
    
    @classmethod
    def help(cls):
        """
        返回当前可用的参数列表。
        """
        return cls.available_init_params()

    def __init__(
        self,
        *,
        name: str = None,
        handlers: List[Union[Callable, Generator, AsyncGenerator]] = None,
        **kwargs
    ):
        """
        Runnable 的构造函数，主要包括：
        - 初始化线程组
        - 工具：作为工具的Runnable列表，在发现工具后是否执行工具的标记等
        """
        raise_invalid_params(kwargs, self.__class__.available_init_params())
        ExecutorManager.__init__(self, **filter_kwargs(kwargs, ExecutorManager.available_init_params()))

        self.name = name or f'{self.__class__.__name__}.{id(self)}'
        self.continue_running = True
        self.handlers = handlers or []
        self.verbose = False

        self._last_output = None

        BindingManager.__init__(self, **filter_kwargs(kwargs, BindingManager.available_init_params()))

    def __repr__(self):
        if self.__class__.__name__ in self.name:
            return f"<{self.name}>"
        else:
            return f"<{self.__class__.__name__} {self.name}>"

    @property
    def selected(self):
        """
        让 Runnable 对象与 Selector 的访问方式兼容。
        """
        return self

    @property
    def last_output(self):
        return self._last_output

    @property
    def provider_dict(self):
        local_dict = {
            "last_output": self.last_output,
        }
        return {
            **super().provider_dict,
            **{k:v for k,v in local_dict.items() if v is not None},
        }

    def __call__(
        self,
        *args,
        verbose: bool = False,
        handlers: List[Union[Callable, Generator, AsyncGenerator]] = None,
        action: str = None,
        **kwargs
    ):
        self.continue_running = True
        handlers = handlers or self.handlers or [log]
        
        try:
            if any(inspect.iscoroutinefunction(handler) for handler in handlers):
                if action:
                    if not hasattr(self, action):
                        raise AttributeError(f"方法 '{action}' 不存在于实例中")
                    method = getattr(self, action)
                else:
                    method = self.async_call

                return self.handle_async_call(*args, verbose=verbose, handlers=handlers, action_method=method, **kwargs)
            else:
                if action:
                    if not hasattr(self, action):
                        raise AttributeError(f"方法 '{action}' 不存在于实例中")
                    method = getattr(self, action)
                else:
                    method = self.call
                return self.handle_sync_call(*args, verbose=verbose, handlers=handlers, action_method=method, **kwargs)
        finally:
            self.continue_running = False

    def handle_sync_call(
        self,
        *args,
        verbose: bool = False,
        handlers: List[Union[Callable, Generator, AsyncGenerator]] = None,
        action_method: Callable = None,
        **kwargs
    ):
        self.verbose = verbose
        if isinstance(handlers, list) and all(callable(handler) for handler in handlers):
            resp = action_method(*args, **kwargs)
            if isinstance(resp, Generator):
                for block in resp:
                    if isinstance(block, str):
                        block = EventBlock("text", block)
                    elif not isinstance(block, EventBlock):
                        block = EventBlock("text", str(block))
                    block.runnable_info = self.runnable_info
                    for handler in handlers:
                        if not inspect.iscoroutinefunction(handler):
                            handler(block, verbose=verbose, **kwargs)
                return self.last_output
            else:
                self._last_output = resp
                return resp
        else:
            raise ValueError("handlers 必须是可调用的列表")

    async def handle_async_call(
        self,
        *args,
        verbose: bool = False,
        handlers: List[Union[Callable, Generator, AsyncGenerator]] = None,
        action_method: Callable = None,
        **kwargs
    ):
        self.verbose = verbose
        if isinstance(handlers, list) and all(callable(handler) for handler in handlers):
            resp = action_method(*args, **kwargs)
            tasks = []
            async def handle_block(block):
                if isinstance(block, str):
                    block = EventBlock("text", block)
                elif not isinstance(block, EventBlock):
                    block = EventBlock("text", str(block))
                block.runnable_info = self.runnable_info
                for handler in handlers:
                    resp = handler(block, verbose=verbose, **kwargs)
                    if inspect.isawaitable(resp):
                        tasks.append(asyncio.create_task(resp))
                if tasks:
                    await asyncio.gather(*tasks)

            if isinstance(resp, AsyncGenerator):
                async for block in resp:
                    await handle_block(block)
            elif isinstance(resp, Generator):
                for block in resp:
                    await handle_block(block)
            else:
                self._last_output = resp

            return self.last_output
        else:
            raise ValueError("handlers 必须是可调用的列表")

        return self.last_output


    @property
    def is_running(self):
        return self.continue_running
    
    @property
    def runnable_info(self):
        return {
            "class_name": self.__class__.__name__,
        }

    def halt(self):
        self.continue_running = False

    @abstractmethod
    def call(self, *args, **kwargs):
        """
        这是同步调用的主入口，默认须实现该方法。
        """
        raise NotImplementedError("子类必须实现 call 方法")

    async def async_call(self, *args, **kwargs):
        """
        默认的异步调用，通过多线程实现。
        请注意，这会制造出大量线程，并不是最佳的性能优化方案。
        虽然不适合大规模部署，但这一方案可以在无需额外开发的情况下支持在异步环境中调用，快速验证业务逻辑。
        """
        loop = asyncio.get_running_loop()
        for block in await self.run_in_executor(self.call, *args, **kwargs):
            yield block

    async def run_in_executor(self, sync_function: Callable, *args, **kwargs):
        loop = asyncio.get_running_loop()
        func = partial(sync_function, *args, **kwargs)
        return await loop.run_in_executor(self.executor, func)

    def _is_running_in_jupyter(self):
        try:
            from IPython import get_ipython
            if 'IPKernelApp' not in get_ipython().config:
                return False
        except Exception:
            return False
        return True

    def bind_consumer(self, runnable, binding_map: Dict=None, dynamic: bool=False):
        """
        传递自身的 handlers 给下游 runnable。
        """
        super().bind_consumer(runnable, binding_map=binding_map, dynamic=dynamic)
        for handler in self.handlers:
            if handler not in runnable.handlers:
                runnable.handlers.append(handler)



import re

from typing import List, Union, Generator, AsyncGenerator
from .....io import EventBlock, NewLineBlock
from .....utils import minify_text, filter_kwargs, raise_invalid_params
from ...selector import Selector
from ..base import BaseAgent

class FlowAgent(BaseAgent):
    @classmethod
    def available_init_params(cls):
        return {
            "max_steps": "最大步骤数",
            **BaseAgent.available_init_params(),
        }

    def __init__(self, *agents, max_steps: int=None, **kwargs):
        raise_invalid_params(kwargs, self.available_init_params())
        super().__init__(**filter_kwargs(kwargs, self.available_init_params()))

        self.max_steps = max_steps or 20

        self.agents = agents
        for r in self.agents:
            if not isinstance(r, (BaseAgent, Selector)):
                raise ValueError("only accept BaseAgent or Selector join to Flow")

    def reset(self):
        pass

    def get_agent_by_name(self, name: str):
        """
        构建一个从名字查找 Agent 和 位置的索引
        """
        all = {a.name: (i, a) for i, a in enumerate(self.agents)}
        return all.get(name, None)

    def call(self, *args, **kwargs):
        """
        执行智能体管道。
        """

        current_agent = self.agents[0]
        current_index = 0
        current_args = args
        current_kwargs = kwargs
        steps_count = 0

        while(steps_count < self.max_steps):
            # 如果 current_agent 是一个选择器
            selected_agent = current_agent.selected
            agent_name = selected_agent if isinstance(selected_agent, str) else selected_agent.name

            # 如果已经到了 __End__  节点，就退出            
            if agent_name.lower() == "__end__":
                break

            # 确保正确获得 agent 和 当前 index
            (current_index, selected_agent) = self.get_agent_by_name(agent_name)

            # 广播节点信息
            info = self._get_node_info(current_index + 1, selected_agent)
            yield EventBlock("agent", info)

            call_resp = selected_agent.call(*current_args, **current_kwargs)
            if isinstance(call_resp, Generator):
                yield from call_resp

            if (current_index + 1) == len(self.agents):
                # 如果已经超出最后一个节点，就结束
                break
            else:
                # 否则继续处理下一个节点
                current_index += 1
                current_agent = self.agents[current_index]

            # 构造下一次调用的参数
            current_args = [selected_agent]
            steps_count += 1

        yield EventBlock("info", f"执行完毕，所有节点运行 {steps_count} 步")

    def _get_node_info(self, index, agent):
        return f">>> Node {index}: {agent.name}"




class History():
    def __init__(self):
        self.history = []

    def append(self, messages: list, **kwargs):
        self.history.append({
            'messages': messages,
            **kwargs
        })
    
    def get(self, index: int=-1):
        return self.history[index]
    


from .block import EventBlock, EndBlock, NewLineBlock, ResponseBlock
from .utils import merge_blocks_by_index
from .log import log, alog
from .usage import usage, async_usage
from .event_stream import event_stream

__all__ = [
    "EventBlock",
    "EndBlock",
    "NewLineBlock",
    "ResponseBlock",
    "log",
    "alog",
    "event_stream",
    "usage",
    "async_usage",
]



from typing import Dict, Any

from ..io import EventBlock
from ..core.runnable.agent import BaseAgent

import textwrap
import pandas as pd
import numpy as np

def parse_code(code: str):
    """
    解析代码，去除代码块的标记。
    """
    return code.replace("```python", "").replace("```", "")

def execute_code(data: Dict[str, Any], code: str):
    """
    执行代码，并返回执行结果。
    """
    filtered_code = '\n'.join([line for line in code.split('\n') if not line.startswith('import')])

    # 受限的全局命名空间
    restricted_globals = {
        "__builtins__": {
            "print": print,
            "range": range,
            "len": len,
            "float": float,
            "int": int,
            "str": str,
            "set": set,
            "dict": dict,
            "list": list,
        },
        "data": data,  # 数据集清单
        "pd": pd  # 仅允许 pandas 模块
    }

    # 将附加代码添加到现有代码中
    filtered_code = f"{filtered_code}\n\nresult = main()\n"

    # 创建一个合并的命名空间
    exec_namespace = restricted_globals.copy()

    try:
        exec(filtered_code, exec_namespace)
    except Exception as e:
        return f"执行代码时发生错误: {e}"

    return exec_namespace.get('result', "生成的代码已经执行，但返回了空结果。")

class PythonCodeTool(BaseAgent):
    def __init__(self, data: Dict[str, "Dataset"], agent: "ChatAgent", name: str=None, description: str=None, template_id: str=None, **kwargs):
        super().__init__(func=self.python_code, **kwargs)

        self.gen_code_agent = agent

        self.data = data

        self.name = name or "python_code"
        self.description = description or f"回答关于[{dataset_names}]等数据集的问题。\n具体包括：{self.__class__.dataset_desc(self.data)}"

    def call(self, question: str, *args, **kwargs):
        dataset_names = ', '.join(self.data.keys())
        dataset_desc = '\n - '.join([ds.desc for ds in self.data.values()])

        output_text = ''
        for block in self.gen_code_agent.call(question, *args, **kwargs):
            if block.block_type == 'chunk':
                output_text += block.text
            yield EventBlock("tool_resp_chunk", block.text)

        code = parse_code(output_text)
        if code:
            exec_result = execute_code(self.data, code)
            yield EventBlock("tool_resp_final", exec_result)
        else:
            yield EventBlock("warn", "没有正确生成python代码失败。")





from typing import List, Union

from .....io import EventBlock
from .....utils import minify_text
from ..base import BaseAgent
from ..chat import ChatAgent

class Pipe(BaseAgent):
    """
    智能体管道，用于将多个智能体节点连接起来，形成一个智能体管道。

    Pipe 是 BaseAgent 的子类，因此可以作为 BaseAgent 使用。
    """
    def __init__(self, *agents):
        """ 
        快速构造智能体运行管道
        """
        for run in agents:
            if not isinstance(run, BaseAgent):
                raise ValueError("agents 必须是 BaseAgent 实例")

        super().__init__("PIPE")
        self.agents = agents

    def __str__(self):
        runnable_types = [type(run).__name__ for run in self.agents]
        return f"Pipe({runnable_types})"

    def __repr__(self):
        runnable_types = [type(run).__name__ for run in self.agents]
        return f"Pipe({runnable_types})"

    def call(self, *args, **kwargs):
        """
        执行智能体管道。
        """
        prev_runnable = None
        for index, run in enumerate(self.agents):
            info = self._get_node_info(index + 1, run)
            yield EventBlock("agent", info)
            if index == 0:
                current_args = args
                current_kwargs = kwargs
            else:
                prompt = prev_runnable.last_output
                current_args = [prompt]
                current_kwargs = {}

            self.create_new_memory(f"节点 <{index}> 正在处理任务...")
            yield from run.call(*current_args, **current_kwargs)
            self.remember_response(run.last_output)
            prev_runnable = run

    def _get_node_info(self, index, run):
        if isinstance(run, ChatAgent):
            if run.memory:
                info = f">>> Node {index}: {minify_text(run.memory[0].get('content'), 30, 30, 10)}"
            else:
                info = f">>> Node {index}: {run.__class__.__name__}"
        else:
            info = f">>> Node {index}: {run.__class__.__name__}"
        return info



import inspect
import json
import textwrap

from typing import Any, Callable, Dict, List

class ToolAbility:
    @classmethod
    def available_init_params(cls):
        """
        返回当前可用的参数列表。
        """
        return {
            "func": "用于自定义工具的同步执行函数",
            "async_func": "用于自定义工具的异步执行函数",
            "name": "工具名称",
            "description": "工具描述",
            "tool_params": "工具参数",
        }

    def __init__(self, *, func: Callable = None, async_func: Callable = None, name: str = None, description: str = None, tool_params: Dict[str, Any] = None):
        self.func = func
        self.async_func = async_func
        self.description = description or "我还没有工具描述"
        self.tool_params = tool_params or {}

        if name:
            self.name = name

    def get_default_parameters(self):
        _parameters = {
            "type": "object",
            "properties": {},
            "required": []
        }
        sig = inspect.signature(self.func or self.async_func or self.call)
        for name, param in sig.parameters.items():
            param_value = param.default if param.default is not inspect.Parameter.empty else ""
            
            # 根据参数的注解类型设置 JSON 兼容类型
            param_type = "string"  # 默认类型
            if param.annotation is int:
                param_type = "integer"
            elif param.annotation is float:
                param_type = "number"
            elif param.annotation is bool:
                param_type = "boolean"
            elif param.annotation is str:
                param_type = "string"
            elif param.annotation is list:
                param_type = "array"
            elif param.annotation is dict:
                param_type = "object"
            
            _parameters["properties"][name] = {
                "type": param_type,
                "description": str(param_value)
            }
            if param.default is inspect.Parameter.empty:
                _parameters["required"].append(name)
        return _parameters

    @property
    def parameters(self) -> str:
        default_params = self.get_default_parameters()
        final_params = {'type': default_params["type"], 'properties': {}, 'required': []}
        if not self.tool_params:
            return default_params
        else:
            for prop_name, prop_dict in default_params["properties"].items():
                if prop_name in self.tool_params.keys():
                    final_params["properties"][prop_name] = prop_dict
                    final_params["properties"][prop_name]["description"] = self.tool_params[prop_name]
            final_params["required"] = [k for k in default_params["required"] if k in final_params["properties"].keys()]
            return final_params

    @property
    def tool_desc(self) -> Dict[str, Any]:
        """
        当作为工具时的自描述信息。

        内容参考了 openai 的工具描述规格生成。
        """
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": self.parameters,
            }
        }

    def parse_arguments(self, arguments: str) -> Dict[str, Any]:
        """
        解析 arguments 字符串并根据 self.parameters 中的类型信息转换为相应的 Python 类型。
        """
        parsed_arguments = json.loads(arguments)
        parameter_types = self.parameters.get("properties", {})

        if isinstance(parsed_arguments, list):
            return parsed_arguments
        elif isinstance(parsed_arguments, dict):
            for key, value in parsed_arguments.items():
                expected_type = parameter_types.get(key, {}).get("type", "string")
                
                if expected_type == "integer":
                    parsed_arguments[key] = int(value)
                elif expected_type == "number":
                    parsed_arguments[key] = float(value)
                elif expected_type == "boolean":
                    if isinstance(value, str):
                        parsed_arguments[key] = value.lower() == "true"
                    else:
                        parsed_arguments[key] = bool(value)
                else:
                    # 默认转换为字符串
                    parsed_arguments[key] = str(value)
            return parsed_arguments
        else:
            return None







from typing import Union, List, Callable

from .....utils import extract_segments, filter_kwargs, raise_invalid_params
from .....io import EventBlock
from ...selector import Selector
from ...prompt_template import PromptTemplate
from ..base import BaseAgent
# from ..tools_calling import BaseToolCalling, Plans
from .base import FlowAgent

class ReWOO(FlowAgent):
    """
    ReWOO 是一种高效的增强语言模型框架，它通过分离 LLM 的推理过程和外部工具调用，并利用可预测推理能力和参数效率，
    实现了更轻量级和可扩展的 ALM 系统。ReWOO 将 ALM 的核心组件分为 Planner、Worker 和 Solver 三个模块，
    Planner 负责制定解决问题的计划，Worker 负责使用外部工具获取证据，Solver 负责根据计划和证据得出最终答案。
    这种方式避免了传统 ALM 中推理和观察的交织，减少了 token 消耗，提高了效率。
    同时，ReWOO 可以通过指令微调和模型专化，将 LLM 的通用推理能力迁移到更小的语言模型中，实现更轻量级的 ALM 系统。
    ReWOO 在多个 NLP 基准数据集上取得了与 ReAct 相当或更好的性能，同时减少了 token 消耗，为构建更高效、可扩展的 ALM 提供了一种新的思路。
    """
    @classmethod
    def available_init_params(cls):
        return {
            "planner": "计划者",
            "solver": "求解者",
            "tools": "工具列表",
            "handler_tool_call": "工具调用处理器",
            **FlowAgent.available_init_params(),
        }
        
    def __init__(
        self,
        planner: BaseAgent=None,
        solver: BaseAgent=None,
        tools: List[BaseAgent]=None,
        handler_tool_call=None,
        **kwargs
    ):
        raise_invalid_params(kwargs, self.available_init_params())

        merged_tools = planner.tools + (tools or [])

        self.planner = planner.reset(
            reinit=True,
            memory=PromptTemplate("FLOW/ReWOO/Planner"),
            tools=merged_tools
        )

        self.solver = solver.reset(
            reinit=True,
            memory=PromptTemplate("FLOW/ReWOO/Solver")
        )

        self.handler_tool_call = handler_tool_call or Plans(tools_to_exec=self.planner.get_tools())

        super().__init__(
            self.planner,
            **filter_kwargs(kwargs, self.available_init_params())
        )

        if not self.planner.get_tools():
            raise ValueError("planner 必须提供 tools")

    def begin_call(self):
        super().begin_call()
        if isinstance(self.handler_tool_call, Plans):
            self.handler_tool_call.reset()

    def after_agent_call(self, agent: BaseAgent):
        """
        在调用完 agent 之后，进行一些处理。
        1. 执行 handler_tool_call.handle(agent.last_output)
        2. 将 handler_tool_call.completed_work 和 提取到 completed_work
        3. 将 task 提交给 solver 合成最后输出
        4. 将 solver 的输出设置为 _last_output 并作为 final_answer 返回
        """
        output = agent.last_output
        self._last_output = agent.provider_dict["task"]

        # 调用工具，并观察工具执行结果
        if self.handler_tool_call:
            for block in self.handler_tool_call.handle(agent.last_output):
                yield block

        # 提取最终答案
        self.solver.reset()
        self.solver.bind_provider(
            binding_map={
                "completed_work": self.handler_tool_call.completed_work
            },
            dynamic=True
        )
        yield from self.solver.call(self.agents[0].provider_dict["task"])

        self.final_answer = self.solver.last_output
        self._last_output = self.final_answer

        yield EventBlock("text", f"\n**最终答案**\n{self.final_answer}")



from .base import (
  load_resource_template,
  load_prompt_template,
  clone_prompt_template,
  get_template_variables
)

__all__ = [
  "load_resource_template",
  "load_prompt_template",
  "clone_prompt_template",
  "get_template_variables",
]


from typing import Union, List, Dict, Any
from ...binding_manager import BindingManager
from ..base import BaseAgent
from .....utils import extract_segments
import json

class BaseTeam(BaseAgent):
    """
    智能体团队通过协作完成任务。
    """
    def __init__(self, leader: BaseAgent, members: List[BaseAgent]=None, **kwargs):
        super().__init__(**kwargs)

        self.members = members or []
        self._completed_teamwork = []

        if not isinstance(leader, BaseAgent):
            raise ValueError("leader must be a BaseAgent instance")

        if len(members) == 0:
            raise ValueError("at least one member must be provided")

        if not all(isinstance(m, BaseAgent) for m in members):
            raise ValueError("members must be a list of BaseAgent instances")

        # 从领导角色和成员角色绑定到 Team
        self.leader.bind_provider(self, {})
        for m in self.members:
            m.bind_provider(self, {})

    @property
    def completed_teamwork(self):
        return self._completed_teamwork
    
    @property
    def provider_dict(self):
        members_desc = []
        names = set([self.leader.name])
        for index, m in enumerate(self.members):
            if m.name not in names:
                _name = m.name
                names.add(_name)
            else:
                _name = f"{m.name}_{index}"
                m.name = _name
            members_desc.append({"成员名字": m.name, "擅长能力": m.description})

        local_dict = {
            "members": json.dumps(members_desc, ensure_ascii=False),
            "completed_teamwork": "\n".join(self.completed_teamwork),
        }
        return {
            **super().provider_dict,
            **{k:v for k,v in local_dict.items() if v is not None},
        }

    def find_member(self, name: str) -> BaseAgent:
        for m in self.members:
            if m.name == name:
                return m
        return None

    def extract_answer(self, text: str) -> str:
        return extract_segments(text, "<final_answer>", "</final_answer>")

    def extract_self_solve(self, text: str) -> str:
        return extract_segments(text, "<self_solve>", "</self_solve>")

    def extract_task_dispatch(self, text: str) -> str:
        tasks = extract_segments(text, "<sub_task>", "</sub_task>")
        for task in tasks:
            try:
                task = json.loads(task)
                tasks.append(task)
            except json.JSONDecodeError:
                pass
        return tasks



from typing import Union, List, Callable

from .....utils import extract_segments, filter_kwargs, raise_invalid_params
from .....io import EventBlock
from ...selector import Selector
from ...prompt_template import PromptTemplate
from ..base import BaseAgent
from .base import FlowAgent

class ReAct(FlowAgent):
    """
    ReAct 提供了一种更易于人类理解、诊断和控制的决策和推理过程。
    它的典型流程可以用一个有趣的循环来描述：思考（Thought）→ 行动（Action）→ 观察（Observation），简称TAO循环。
    """
    @classmethod
    def available_init_params(cls):
        return {
            "planner": "用于推理的ChatAgent, 你应当在其中指定可用的工具",
            "prompt_template": "用于生成计划的PromptTemplate, 默认为 PromptTemplate('FLOW/ReAct/Planner')",
            "final_answer_prompt": "最终答案提示词关键字, 默认为 **最终答案**",
            **FlowAgent.available_init_params(),
        }

    def __init__(
        self,
        planner: BaseAgent,
        prompt_template: str=None,
        final_answer_prompt: str=None,
        **kwargs
    ):
        raise_invalid_params(kwargs, self.available_init_params())

        if not isinstance(planner, BaseAgent):
            raise ValueError("planner 必须是 ChatAgent 的子类")
        if not planner.tools:
            raise ValueError("planner 必须包含可用的工具")

        final_answer_prompt = final_answer_prompt or "**最终答案**"
        # 设置 planner 的 tools_behavior 为 "parse-execute"，执行后就停止，等待 T-A-O 循环
        planner.tools_behavior = "parse-execute"
        prompt_template = prompt_template or PromptTemplate("FLOW/ReAct/Planner")
        planner.set_init_messages(prompt_template)

        class Observer(BaseAgent):
            def __init__(self, **kwargs):
                super().__init__(**kwargs)
                self.completed_work = []
                self.final_answer = None

                # 将 Planner 绑定给 Observer
                # 主要获得 task, tools_calling_steps
                self.bind_provider(planner)

                # 将 Observer 绑定到 prompt_template
                # 获得 task, completed_work
                self.bind_consumer(prompt_template)
            
            @property
            def provider_dict(self):
                return {
                    "completed_work": "\n".join(self.completed_work),
                    **super().provider_dict,
                }

            def call(self, agent: BaseAgent, **kwargs):
                # 提取最终答案
                output = agent.last_output

                if output:
                    self.completed_work.append(output)
                else:
                    self.final_answer = "没有可用的输出"
                    yield EventBlock("warn", f"{final_answer_prompt}\n{self.final_answer}")
                    return

                if final_answer_prompt in output:
                    final_answer_index = output.index(final_answer_prompt)
                    self.final_answer = output[final_answer_index:].split(final_answer_prompt)[-1].strip()
                    yield EventBlock("text", f"\n**最终答案**\n{self.final_answer}\n")

                if agent.tools_calling_steps:
                    all_results = "\n".join([step["result"] for step in agent.tools_calling_steps])
                    observation = f'\n**观察**\n上面的行动结果为:\n{all_results}\n'

                    self.completed_work.append(observation)
                    self._last_output = self.consumer_dict.get("task", "请开始")

                    yield EventBlock("text", observation)

        observer = Observer(name="observer")

        def should_continue(vars, runs):
            if (final_answer_prompt in planner.last_output) or observer.final_answer:
                return "END"
            else:
                planner.memory.clear()
                return planner

        super().__init__(
            planner,
            observer,
            Selector([], condition=should_continue, name="react_selector"),
            **filter_kwargs(kwargs, self.available_init_params())
        )



from typing import List, Union

from .....io import EventBlock
from ....markdown import Markdown
from ..base import BaseAgent
from ..chat import ChatAgent
from .pipe import Pipe

import copy

class FromOutline(BaseAgent):
    """
    实现扩写：从输出结果提炼大纲，然后生成内容。
    """
    def __init__(self, writer: ChatAgent=None, prev_k:int=1000, next_k:int=500):
        if not isinstance(writer, ChatAgent):
            raise ValueError("扩写智能体 writer 必须是 ChatAgent 实例")

        super().__init__("FROM_OUTLINE")
        self.writer = writer
        self.prev_k = prev_k
        self.next_k = next_k

        # 执行 call 调用时提供
        self.outline_text = None

        # 执行 call 调用时生成
        self.markdown = Markdown()
        self.outline = []
        self.agents = {}

    @property
    def last_output(self):
        """
        从 Markdown 生成当前的文本输出内容。
        如果当前有提纲，则将提纲中的内容替换为实际内容。
        """
        if self.outline:
            md = copy.deepcopy(self.markdown)
            for doc in self.outline:
                if doc.meta['id'] in self.agents:
                    from_outline_text = self.agents[doc.meta['id']].memory[-1]['content']
                    md.replace_documents(doc, doc, from_outline_text)
            return md.text
        else:
            return self.outline_text

    def call(self, outline_text: str, *args, **kwargs):
        """
        执行扩写。
        """

        if not isinstance(outline_text, str):
            raise ValueError("outline_text 必须是字符串")

        if self.outline_text != outline_text:
            self.markdown = Markdown(outline_text)
            self.outline = self.markdown.get_outline() if self.markdown else []
            self.outline_text = outline_text

        if self.outline:
            for doc in self.outline:
                outline_id = doc.meta['id']

                segment_writer = self.writer.clone()
                self.agents[doc.meta['id']] = segment_writer

                (draft, outline) = self.markdown.fetch_outline_task(doc, prev_k=self.prev_k, next_k=self.next_k)
                segment_writer.set_outline(f'```markdown\n{outline}\n```')
                segment_writer.set_draft(f'```markdown\n{draft}\n```')

                info = f"执行扩写任务 <{outline_id}>：\n{outline}"
                yield EventBlock("agent", info)
                self.create_new_memory(info)
                yield from segment_writer.call("请开始扩写")
                self.remember_response(segment_writer.last_output)
        else:
            yield EventBlock("info", f"没有提纲可供扩写")




from .base import Runnable, BaseAgent
from .chat import ChatAgent, ChatPool
from .chat import BaseToolCalling, ToolCall, SubTask, Plans
from .team import BaseTeam, StepByStep, Pipe, FromOutline, Discuss
from .retriever import Retriever
from .flow import FlowAgent, ReAct, ReWOO



import re
from typing import Any, Set, Union, List
from ...document import Document
from ..vectordb import VectorDB

class KnowledgeManager:
    @classmethod
    def available_init_params(cls):
        """
        返回当前可用的参数列表。
        """
        return {
            "knowledge": "知识列表",
        }

    def __init__(self, knowledge: Union[Set[Any], List[Any]] = None):
        """
        知识库在内存中以集合的方式保存，确保了其具有唯一性。
        """
        if isinstance(knowledge, set):
            for text in knowledge:
                if not isinstance(item, (str, Document, VectorDB)):
                    raise ValueError("Knowledge list items must be str, Document, or VectorDB")
            self.knowledge = knowledge
        else:
            self.knowledge = set(knowledge or {})

    def add_knowledge(self, text: str):
        if isinstance(text, str):
            self.knowledge.add(Document(text))
        elif isinstance(text, Document):
            self.knowledge.add(text)
        else:
            raise ValueError("Knowledge must be a string or a Document")

    def get_knowledge(self, query: str=None, verbose: bool=False):
        knowledge = []
        for kg in self.knowledge:
            if isinstance(kg, Document):
                knowledge.append(kg.text)
            elif isinstance(kg, str):
                knowledge.append(kg)
            elif isinstance(kg, VectorDB):
                query_results = [doc.text for doc in kg(query, verbose=verbose)]
                knowledge.append("\n\n".join(query_results))
            else:
                raise ValueError("Knowledge must be a string, Document or VectorDB")
        return knowledge

    def clear_knowledge(self):
        self.knowledge.clear()
